{%- if prev %}
  <li><a href="{{ prev.link|e }}"
         title="{{ _('previous chapter') }}">{{ "&laquo;"|safe }} Previous</a></li>
{%- endif %}
{%- if next %}
  <li><a href="{{ next.link|e }}"
         title="{{ _('next chapter') }}">Next {{ "&raquo;"|safe }}</a></li>
{%- endif %}
