// Various mathematical/computational constants

#if !defined(CU_CONSTANTS_H_)
#define CU_CONSTANTS_H_

#define PI   3.14159265358979323846               // pi
#define PIO2 1.57079632679489661923               // pi/2
#define EUL  0.577215664901532860606512090082402  // Euler's constant

// maximum single and double precision floating point numbers
#ifndef FLT_MAX
#define FLT_MAX 3.402823466E+38F
#endif

#ifndef DBL_MAX
#define DBL_MAX 1.7976931348623158e+308
#endif

#endif /* !defined(CU_CONSTANTS_H_) */
